#  BSD 3-Clause License
#
#  Copyright (c) 2012, the Sentry Team, see AUTHORS for more details
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE


from __future__ import absolute_import

import django
from django.conf import settings as django_settings
from django.db import DatabaseError
from django.http import HttpRequest

try:
    from rest_framework.request import Request as DrfRequest
except ImportError:
    DrfRequest = HttpRequest

from atatus import get_client as _get_client
from atatus.base import Client
from atatus.conf import constants
from atatus.contrib.django.utils import get_raw_uri, iterate_with_template_sources
from atatus.utils import compat, encoding, get_url_dict
from atatus.utils.encoding import long_field
from atatus.utils.logging import get_logger
from atatus.utils.module_import import import_string
from atatus.utils.wsgi import get_environ, get_headers

__all__ = ("DjangoClient",)


default_client_class = "atatus.contrib.django.DjangoClient"


def get_client():
    """
    Get an Atatus client.

    :param client:
    :return:
    :rtype: atatus.base.Client
    """
    c = _get_client()
    if c:
        return c

    config = getattr(django_settings, "ATATUS", {})
    client = config.get("CLIENT", default_client_class)
    client_class = import_string(client)
    instance = client_class()
    # `instance` will already be in atatus.base.CLIENT_SINGLETON due to the
    # `__init__()` for Client
    return instance


class DjangoClient(Client):
    logger = get_logger("atatus.errors.client.django")

    def __init__(self, config=None, **inline) -> None:
        if config is None:
            config = getattr(django_settings, "ATATUS", {})
        if "framework_name" not in inline:
            inline["framework_name"] = "Django"
            inline["framework_version"] = django.get_version()
        super(DjangoClient, self).__init__(config, **inline)

    def get_user_info(self, request):
        user_info = {}

        if not hasattr(request, "user"):
            return user_info
        try:
            user = request.user
            if hasattr(user, "is_authenticated"):
                if callable(user.is_authenticated):
                    user_info["is_authenticated"] = user.is_authenticated()
                else:
                    user_info["is_authenticated"] = bool(user.is_authenticated)
            if hasattr(user, "id"):
                user_info["id"] = encoding.keyword_field(user.id)
            if hasattr(user, "get_username"):
                user_info["username"] = encoding.keyword_field(encoding.force_text(user.get_username()))
            elif hasattr(user, "username"):
                user_info["username"] = encoding.keyword_field(encoding.force_text(user.username))

            if hasattr(user, "email"):
                user_info["email"] = encoding.force_text(user.email)
        except DatabaseError:
            # If the connection is closed or similar, we'll just skip this
            return {}

        return user_info

    def get_data_from_request(self, request, event_type):
        result = {
            "env": dict(get_environ(request.META)),
            "method": request.method,
            "socket": {"remote_address": request.META.get("REMOTE_ADDR")},
            "cookies": dict(request.COOKIES),
        }
        if self.config.capture_headers:
            request_headers = dict(get_headers(request.META))

            for key, value in request_headers.items():
                if isinstance(value, (int, float)):
                    request_headers[key] = str(value)

            result["headers"] = request_headers

        if request.method in constants.HTTP_WITH_BODY:
            capture_body = self.config.capture_body in ("all", "request")
            if not capture_body:
                result["body"] = "[REDACTED]"
            else:
                content_type = request.META.get("CONTENT_TYPE")
                if content_type and content_type.startswith("multipart/form-data"):
                    data = compat.multidict_to_dict(request.POST)
                    if request.FILES:
                        data["_files"] = {field: file.name for field, file in request.FILES.items()}
                else:
                    try:
                        if isinstance(request._atatus_body, bytes):
                            data = encoding.body_decode(request._atatus_body)
                        else:
                            data = request._atatus_body
                    except Exception as e:
                        self.logger.debug("Can't capture request body: %s", str(e))
                        data = "<unavailable>"
                if data is not None:
                    # Can we apply this as a processor instead?
                    # https://github.com/elastic/apm-agent-python/issues/305
                    result["body"] = long_field(data)

        url = get_raw_uri(request)
        try:
            result["url"] = get_url_dict(url)
        except ValueError as exc:
            self.logger.warning(f"URL parsing failed: {exc}")
        return result



    def get_data_from_response(self, response, event_type):
        result = {"status_code": response.status_code}

        if self.config.capture_headers and hasattr(response, "items"):
            response_headers = dict(response.items())

            for key, value in response_headers.items():
                if isinstance(value, (int, float)):
                    response_headers[key] = str(value)

            result["headers"] = response_headers

        capture_body = self.config.capture_body in ("all", "response")
        if capture_body and self.config.analytics and hasattr(response, "content"):
            captureContent = False
            if "headers" in result and "Content-Type" in result["headers"]:
                contentType = result["headers"]["Content-Type"]
                for ct in self.config.log_body_content_types:
                    if ct in contentType:
                        captureContent = True
                        break

            if captureContent is True:
                content_encoding = ""
                if "headers" in result and "Content-Encoding" in result["headers"] and result["headers"]["Content-Encoding"] is not None and "gzip" in (result["headers"]["Content-Encoding"]).lower():
                    content_encoding = "gzip"

                body = None
                if isinstance(response.content, str):
                    body = encoding.body_string(response.content, content_encoding)
                else:
                    body = encoding.body_bytes(response.content, content_encoding)
                if body is not None:
                    decoded_body = encoding.body_decode(body)
                    result["body"] = encoding.body_field(str(decoded_body))


        return result

    def capture(self, event_type, request=None, **kwargs):
        if "context" not in kwargs:
            kwargs["context"] = context = {}
        else:
            context = kwargs["context"]

        is_http_request = isinstance(request, (HttpRequest, DrfRequest))
        if is_http_request:
            context["request"] = self.get_data_from_request(request, constants.ERROR)
            context["user"] = self.get_user_info(request)

        result = super(DjangoClient, self).capture(event_type, **kwargs)

        if is_http_request:
            # attach the atatus object to the request
            request._atatus = {"service_name": self.config.service_name, "id": result}

        return result

    def _get_stack_info_for_trace(
        self,
        frames,
        library_frame_context_lines=None,
        in_app_frame_context_lines=None,
        with_locals=True,
        locals_processor_func=None,
    ):
        """If the stacktrace originates within the atatus module, it will skip
        frames until some other module comes up."""
        return list(
            iterate_with_template_sources(
                frames,
                with_locals=with_locals,
                library_frame_context_lines=library_frame_context_lines,
                in_app_frame_context_lines=in_app_frame_context_lines,
                include_paths_re=self.include_paths_re,
                exclude_paths_re=self.exclude_paths_re,
                locals_processor_func=locals_processor_func,
            )
        )

    def send(self, url, **kwargs):
        """
        Serializes and signs ``data`` and passes the payload off to ``send_remote``

        If ``server`` was passed into the constructor, this will serialize the data and pipe it to
        the server using ``send_remote()``.
        """
        if self.config.server_url:
            return super(DjangoClient, self).send(url, **kwargs)
        else:
            self.error_logger.error("No server configured, and atatus not installed. Cannot send message")
            return None


class ProxyClient(object):
    """
    A proxy which represents the current client at all times.
    """

    # introspection support:
    __members__ = property(lambda x: x.__dir__())

    # Need to pretend to be the wrapped class, for the sake of objects that care
    # about this (especially in equality tests)
    __class__ = property(lambda x: get_client().__class__)

    __dict__ = property(lambda o: get_client().__dict__)

    __repr__ = lambda: repr(get_client())
    __getattr__ = lambda x, o: getattr(get_client(), o)
    __setattr__ = lambda x, o, v: setattr(get_client(), o, v)
    __delattr__ = lambda x, o: delattr(get_client(), o)

    __lt__ = lambda x, o: get_client() < o
    __le__ = lambda x, o: get_client() <= o
    __eq__ = lambda x, o: get_client() == o
    __ne__ = lambda x, o: get_client() != o
    __gt__ = lambda x, o: get_client() > o
    __ge__ = lambda x, o: get_client() >= o
    __hash__ = lambda x: hash(get_client())
    # attributes are currently not callable
    # __call__ = lambda x, *a, **kw: get_client()(*a, **kw)
    __nonzero__ = lambda x: bool(get_client())
    __len__ = lambda x: len(get_client())
    __getitem__ = lambda x, i: get_client()[i]
    __iter__ = lambda x: iter(get_client())
    __contains__ = lambda x, i: i in get_client()
    __getslice__ = lambda x, i, j: get_client()[i:j]
    __add__ = lambda x, o: get_client() + o
    __sub__ = lambda x, o: get_client() - o
    __mul__ = lambda x, o: get_client() * o
    __floordiv__ = lambda x, o: get_client() // o
    __mod__ = lambda x, o: get_client() % o
    __divmod__ = lambda x, o: get_client().__divmod__(o)
    __pow__ = lambda x, o: get_client() ** o
    __lshift__ = lambda x, o: get_client() << o
    __rshift__ = lambda x, o: get_client() >> o
    __and__ = lambda x, o: get_client() & o
    __xor__ = lambda x, o: get_client() ^ o
    __or__ = lambda x, o: get_client() | o
    __div__ = lambda x, o: get_client().__div__(o)
    __truediv__ = lambda x, o: get_client().__truediv__(o)
    __neg__ = lambda x: -(get_client())
    __pos__ = lambda x: +(get_client())
    __abs__ = lambda x: abs(get_client())
    __invert__ = lambda x: ~(get_client())
    __complex__ = lambda x: complex(get_client())
    __int__ = lambda x: int(get_client())
    __float__ = lambda x: float(get_client())
    __str__ = lambda x: str(get_client())
    __unicode__ = lambda x: str(get_client())
    __oct__ = lambda x: oct(get_client())
    __hex__ = lambda x: hex(get_client())
    __index__ = lambda x: get_client().__index__()
    __coerce__ = lambda x, o: x.__coerce__(x, o)
    __enter__ = lambda x: x.__enter__()
    __exit__ = lambda x, *a, **kw: x.__exit__(*a, **kw)


client = ProxyClient()


def _get_installed_apps_paths():
    """
    Generate a list of modules in settings.INSTALLED_APPS.
    """
    out = set()
    for app in django_settings.INSTALLED_APPS:
        out.add(app)
    return out



from typing import Callable, Any, Optional
from atatus.base import get_client
from atatus.instrumentation.packages.base import AbstractInstrumentedModule
from atatus.utils.logging import get_logger

logger = get_logger("atatus")

class OtelTracerInstrumentation(AbstractInstrumentedModule):
    name = "opentelemetry-tracer"
    atatus_get_tracer: Optional[Callable[..., Any]]
    instrument_list = ()

    try:
        from atatus.contrib.opentelemetry.trace import get_tracer
        atatus_get_tracer = get_tracer
        instrument_list = (("opentelemetry.trace", "get_tracer"),)
    except ImportError:
        logger.debug("OpenTelemetry not installed, skipping OtelTracerInstrumentation.")
        atatus_get_tracer = None

    creates_transactions = True

    def call(self, module, method, wrapped, instance, args, kwargs):
        if self.atatus_get_tracer is None:
            return wrapped(*args, **kwargs)
        kwargs['atatus_client'] = get_client()
        return self.atatus_get_tracer(*args, **kwargs)



#  BSD 3-Clause License
#
#  Copyright (c) 2023, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
from importlib import import_module

import atatus
from atatus import Client, get_client
from atatus.contrib.serverless.aws import _lambda_transaction, prep_kwargs
from atatus.utils.logging import get_logger

logger = get_logger("atatus.lambda")

# Prep client and instrument
# For some reason, if we instrument as part of our handler, it adds 3+ seconds
# to the cold start time. So we do it here. I still don't know why this slowdown
# happens.
client_kwargs = prep_kwargs()
client = get_client()
if not client:
    client = Client(**client_kwargs)
client.activation_method = "wrapper"
if not client.config.debug and client.config.instrument and client.config.enabled:
    atatus.instrument()


class LambdaError(Exception):
    pass


def lambda_handler(event, context):
    """
    This handler is designed to replace the default handler in the lambda
    function, and then call the actual handler, which will be stored in
    ELASTICAPM_ORIGINAL_HANDLER.
    """
    # Prep original handler
    original_handler = os.environ.get("ELASTICAPM_ORIGINAL_HANDLER", None)
    if not original_handler:
        raise LambdaError("Cannot find original handler. ELASTICAPM_ORIGINAL_HANDLER is not set correctly.")
    try:
        module, handler = original_handler.rsplit(".", 1)
    except ValueError:
        raise LambdaError(f"ELASTICAPM_ORIGINAL_HANDLER is not set correctly: {original_handler}")

    # Import handler
    module = import_module(module.replace("/", "."))
    wrapped = getattr(module, handler)

    client = get_client()

    # Run the handler
    if not client.config.debug and client.config.instrument and client.config.enabled:
        with _lambda_transaction(wrapped, None, client, event, context) as sls:
            sls.response = wrapped(event, context)
            return sls.response
    else:
        return wrapped(event, context)



#  BSD 3-Clause License
#
#  Copyright (c) 2022, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import os
import shutil

import atatus


def setup() -> None:
    parser = argparse.ArgumentParser(
        prog="atatus-run",
        description="""
        %(prog)s is a wrapper script for running python applications
        while automatically instrumenting with the Atatus python agent.
        """,
    )

    parser.add_argument(
        "--version", help="Print Atatus version", action="version", version="%(prog)s " + atatus.VERSION
    )

    parser.add_argument(
        "--config",
        action="append",
        help="Config values to pass to Atatus. Can be used multiple times. Ex: --config 'service_name=foo'",
    )
    parser.add_argument("app", help="Your python application")
    parser.add_argument("app_args", nargs=argparse.REMAINDER, help="Arguments for your python application", default=[])

    args = parser.parse_args()

    our_path = os.path.dirname(os.path.abspath(__file__))
    pythonpath = os.environ.get("PYTHONPATH", "").split(";")
    pythonpath = [path for path in pythonpath if path != our_path]
    pythonpath.insert(0, our_path)
    os.environ["PYTHONPATH"] = os.path.pathsep.join(pythonpath)

    for config in args.config or []:
        key, value = config.split("=", 1)
        os.environ["ATATUS_" + key.upper()] = value

    if args.app:
        app = shutil.which(args.app)
        os.execl(app, app, *args.app_args)



class Layer(object):
    kinds_dict = {
        'db': 'Database',
        'cache': 'Database',
        'external': 'Remote',
        'aws': 'Remote',
        'websocket': 'Remote',
        'template': 'Template',
        'compression': 'Compression',
        'code.custom': 'Custom'
    }

    types_dict = {
        'mysql': 'MySQL',
        'mysql2': 'MySQL',
        'postgresql': 'Postgres',
        'mssql': 'MS SQL',
        'mongodb': 'MongoDB',
        'memcached': 'Memcached',
        'redis': 'Redis',
        'graphql': 'GraphQL',
        'elasticsearch': 'Elasticsearch',
        'cassandra': 'Cassandra',
        'sqlite': 'SQLite',
        'http': 'External Requests',
        'django': 'Django',
        'jinja2': 'Jinja 2',
        'zlib': 'zlib'
    }

    def __init__(self, type, kind, duration):
        self.type = type
        self.kind = kind
        self.count = 1
        self.min = duration
        self.max = duration
        self.total = duration

    def aggregate(self, duration):
        self.count += 1
        if duration < self.min:
            self.min = duration
        if duration > self.max:
            self.max = duration
        self.total += duration



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import functools
from types import TracebackType
from typing import Any, Awaitable, Callable, Optional, Type, TypeVar

from atatus.conf.constants import LABEL_RE
from atatus.traces import SpanType, capture_span, execution_context
from atatus.utils import get_name_from_func

FuncType = Callable[..., Awaitable[Any]]
_AnnotatedFunctionT = TypeVar("_AnnotatedFunctionT", bound=FuncType)


class async_capture_span(capture_span):
    def __call__(self, func: _AnnotatedFunctionT) -> _AnnotatedFunctionT:
        self.name = self.name or get_name_from_func(func)

        @functools.wraps(func)
        async def decorated(*args, **kwds):
            async with self:
                return await func(*args, **kwds)

        return decorated

    async def __aenter__(self) -> Optional[SpanType]:
        return self.handle_enter(False)

    async def __aexit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        self.handle_exit(exc_type, exc_val, exc_tb)


async def set_context(data, key="custom") -> None:
    """
    Asynchronous copy of atatus.traces.set_context().
    Attach contextual data to the current transaction and errors that happen during the current transaction.

    If the transaction is not sampled, this function becomes a no-op.

    :param data: a dictionary, or a callable that returns a dictionary
    :param key: the namespace for this data
    """
    transaction = execution_context.get_transaction()
    if not (transaction and transaction.is_sampled):
        return
    if callable(data):
        data = await data()

    # remove invalid characters from key names
    for k in list(data.keys()):
        if LABEL_RE.search(k):
            data[LABEL_RE.sub("_", k)] = data.pop(k)

    if key in transaction.context:
        transaction.context[key].update(data)
    else:
        transaction.context[key] = data



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Provides classes to instrument dbapi2 providers

https://www.python.org/dev/peps/pep-0249/
"""

import wrapt

from atatus.contrib.asyncio.traces import async_capture_span
from atatus.instrumentation.packages.asyncio.base import AsyncAbstractInstrumentedModule
from atatus.instrumentation.packages.dbapi2 import EXEC_ACTION, QUERY_ACTION
from atatus.utils.encoding import shorten


class AsyncCursorProxy(wrapt.ObjectProxy):
    provider_name = None
    DML_QUERIES = ("INSERT", "DELETE", "UPDATE")

    def __init__(self, wrapped, destination_info=None):
        super(AsyncCursorProxy, self).__init__(wrapped)
        self._self_destination_info = destination_info or {}

    async def callproc(self, procname, params=None):
        return await self._trace_sql(self.__wrapped__.callproc, procname, params, action=EXEC_ACTION)

    async def execute(self, sql, params=None):
        return await self._trace_sql(self.__wrapped__.execute, sql, params)

    async def executemany(self, sql, param_list):
        return await self._trace_sql(self.__wrapped__.executemany, sql, param_list)

    def _bake_sql(self, sql):
        """
        Method to turn the "sql" argument into a string. Most database backends simply return
        the given object, as it is already a string
        """
        return sql

    async def _trace_sql(self, method, sql, params, action=QUERY_ACTION):
        sql_string = self._bake_sql(sql)
        if action == EXEC_ACTION:
            signature = sql_string + "()"
        else:
            signature = self.extract_signature(sql_string)

        # Truncate sql_string to 10000 characters to prevent large queries from
        # causing an error to APM server.
        sql_string = shorten(sql_string, string_length=10000)

        async with async_capture_span(
            signature,
            span_type="db",
            span_subtype=self.provider_name,
            span_action=action,
            extra={
                "db": {"type": "sql", "statement": sql_string, "instance": getattr(self, "_self_database", None)},
                "destination": self._self_destination_info,
            },
            skip_frames=1,
            leaf=True,
        ) as span:
            if params is None:
                result = await method(sql)
            else:
                result = await method(sql, params)
            # store "rows affected", but only for DML queries like insert/update/delete
            if span and self.rowcount not in (-1, None) and signature.startswith(self.DML_QUERIES):
                span.update_context("db", {"rows_affected": self.rowcount})
            return result

    def extract_signature(self, sql):
        raise NotImplementedError()


class AsyncConnectionProxy(wrapt.ObjectProxy):
    cursor_proxy = AsyncCursorProxy

    def __init__(self, wrapped, destination_info=None):
        super(AsyncConnectionProxy, self).__init__(wrapped)
        self._self_destination_info = destination_info

    def cursor(self, *args, **kwargs):
        return self.cursor_proxy(self.__wrapped__.cursor(*args, **kwargs), self._self_destination_info)


class AsyncDbApi2Instrumentation(AsyncAbstractInstrumentedModule):
    connect_method = None

    async def call(self, module, method, wrapped, instance, args, kwargs):
        return AsyncConnectionProxy(await wrapped(*args, **kwargs))

    async def call_if_sampling(self, module, method, wrapped, instance, args, kwargs):
        # Contrasting to the superclass implementation, we *always* want to
        # return a proxied connection, even if there is no ongoing atatus
        # transaction yet. This ensures that we instrument the cursor once
        # the transaction started.
        return await self.call(module, method, wrapped, instance, args, kwargs)



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from atatus.instrumentation.packages.base import AbstractInstrumentedModule
from atatus.traces import capture_span
from atatus.traces import execution_context
from atatus.utils import get_host_from_url, sanitize_url, get_external_headers, external_body_decode
from atatus.utils.encoding import long_field
from atatus.conf import constants


class HttpxClientInstrumentation(AbstractInstrumentedModule):
    name = "httpx"

    instrument_list = [("httpx", "Client.send")]

    def call(self, module, method, wrapped, instance, args, kwargs):

        request = kwargs.get("request") or args[0]

        capture_req_body = False
        capture_res_body = False
        analytics = False
        analytics_capture_outgoing = False

        transaction = execution_context.get_transaction()
        log_body_content_types = []
        if transaction:
            analytics = transaction.tracer.config.analytics
            log_body_content_types = transaction.tracer.config.log_body_content_types
            analytics_capture_outgoing = transaction.tracer.config.analytics_capture_outgoing
            if analytics is True and analytics_capture_outgoing is True:
                capture_req_body = transaction.tracer.config.capture_body in ("all", "request")
                capture_res_body = transaction.tracer.config.capture_body in ("all", "response")

        req_method = request.method.upper()
        if req_method not in constants.HTTP_WITH_BODY:
            capture_req_body = False
        signature = req_method
        signature += " " + get_host_from_url(str(request.url))
        url = sanitize_url(str(request.url))


        req_headers = {}
        req_body = ""
        if analytics is True and analytics_capture_outgoing is True:
            req_headers = {}
            if request.headers:
                req_headers = get_external_headers(dict(request.headers))

            if capture_req_body is True:
                if request.content:
                    data = ""
                    try:
                        if isinstance(request.content, bytes):
                            data = external_body_decode(request.content)
                        else:
                            data = request.content
                    except Exception as e:
                        data = "<unavailable>"

                    if data is not None:
                        # Can we apply this as a processor instead?
                        # https://github.com/elastic/apm-agent-python/issues/305
                        req_body = long_field(data)

        with capture_span(
            signature,
            span_type="external",
            span_subtype="http",
            extra={"http": {"url": url}},
            leaf=True,
        ) as span:
            response = wrapped(*args, **kwargs)
            if response is not None:
                res_headers = {}
                res_body = ""

                captureContent = False
                if analytics is True and analytics_capture_outgoing is True:
                    if response.headers:
                        res_headers = get_external_headers(response.headers)
                    if "content-type" in res_headers:
                        contentType = res_headers["content-type"]
                        for ct in log_body_content_types:
                            if ct in contentType:
                                captureContent = True
                                break
                    if capture_res_body is True and captureContent is True:
                        if response.content:
                            data = ""
                            try:
                                if isinstance(response.content, bytes):
                                    data = external_body_decode(response.content)
                                else:
                                    data = response.content
                            except Exception as e:
                                data = "<unavailable>"
                            if data is not None:
                                # Can we apply this as a processor instead?
                                # https://github.com/elastic/apm-agent-python/issues/305
                                res_body = long_field(data)

                if span.context:
                    span.context["http"]["status_code"] = response.status_code
                    if analytics is True and analytics_capture_outgoing is True:
                        span.context["analytics_captured"] = True

                        span.context["http"]["request"] = {}
                        span.context["http"]["request"]["method"] = req_method
                        span.context["http"]["request"]["headers"] = req_headers
                        if capture_req_body is True:
                            span.context["http"]["request"]["body"] = req_body
                    
                        span.context["http"]["response"] = {}
                        span.context["http"]["response"]["headers"] = res_headers
                        if capture_res_body is True and captureContent is True:
                            span.context["http"]["response"]["body"] = res_body

                span.set_success() if response.status_code < 400 else span.set_failure()
            return response



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import atatus
from atatus.instrumentation.packages.asyncio.base import AsyncAbstractInstrumentedModule
from atatus.instrumentation.packages.elasticsearch import (
    ElasticsearchConnectionInstrumentation,
    ElasticsearchTransportInstrumentation,
)
from atatus.traces import DroppedSpan, execution_context


class ElasticSearchAsyncConnection(ElasticsearchConnectionInstrumentation, AsyncAbstractInstrumentedModule):
    name = "elasticsearch_connection"

    def get_instrument_list(self):
        try:
            import elastic_transport  # noqa: F401

            return [
                ("elastic_transport._node._http_aiohttp", "AiohttpHttpNode.perform_request"),
            ]
        except ImportError:
            return [
                ("elasticsearch_async.connection", "AIOHttpConnection.perform_request"),
                ("elasticsearch._async.http_aiohttp", "AIOHttpConnection.perform_request"),
            ]

    async def call(self, module, method, wrapped, instance, args, kwargs):
        span = execution_context.get_span()
        if not span or isinstance(span, DroppedSpan):
            return await wrapped(*args, **kwargs)

        self._update_context_by_request_data(span.context, instance, args, kwargs)

        result = await wrapped(*args, **kwargs)
        if hasattr(result, "meta"):  # elasticsearch-py 8.x+
            status_code = result.meta.status
            cluster = result.meta.headers.get("x-found-handling-cluster")
        else:
            status_code = result[0]
            cluster = result[1].get("x-found-handling-cluster")

        span.update_context("http", {"status_code": status_code})
        if cluster:
            span.update_context("db", {"instance": cluster})

        return result


class ElasticsearchAsyncTransportInstrumentation(
    ElasticsearchTransportInstrumentation, AsyncAbstractInstrumentedModule
):
    name = "elasticsearch_connection"

    instrument_list = [
        ("elasticsearch._async.transport", "AsyncTransport.perform_request"),
    ]

    def get_instrument_list(self):
        try:
            import elastic_transport  # noqa: F401

            return [
                ("elastic_transport", "AsyncTransport.perform_request"),
            ]
        except ImportError:
            return [
                ("elasticsearch._async.transport", "AsyncTransport.perform_request"),
            ]

    async def call(self, module, method, wrapped, instance, args, kwargs):
        async with atatus.async_capture_span(
            self._get_signature(args, kwargs),
            span_type="db",
            span_subtype="elasticsearch",
            span_action="query",
            extra={},
            skip_frames=2,
            leaf=True,
        ) as span:
            result_data = await wrapped(*args, **kwargs)

            hits = self._get_hits(result_data)
            if hits:
                span.update_context("db", {"rows_affected": hits})

            return result_data



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import absolute_import

from typing import Optional, Union

from atatus.instrumentation.packages.dbapi2 import (
    ConnectionProxy,
    CursorProxy,
    DbApi2Instrumentation,
    extract_signature,
)
from atatus.traces import capture_span
from atatus.utils import default_ports


class PGCursorProxy(CursorProxy):
    provider_name = "postgresql"

    def _bake_sql(self, sql):
        from psycopg2 import extensions as psycopg2_extensions

        # if this is a Composable object, use its `as_string` method
        # see http://initd.org/psycopg/docs/sql.html
        if hasattr(sql, "as_string"):
            sql = sql.as_string(self.__wrapped__)
        # if the sql string is already a byte string, we need to decode it using the connection encoding
        if isinstance(sql, bytes):
            sql = sql.decode(psycopg2_extensions.encodings[self.__wrapped__.connection.encoding])
        return sql

    def extract_signature(self, sql):
        return extract_signature(sql)

    def __enter__(self):
        return PGCursorProxy(self.__wrapped__.__enter__(), destination_info=self._self_destination_info)

    @property
    def _self_database(self):
        return self.connection.info.dbname or ""


class PGConnectionProxy(ConnectionProxy):
    cursor_proxy = PGCursorProxy

    def __enter__(self):
        return PGConnectionProxy(self.__wrapped__.__enter__(), destination_info=self._self_destination_info)


class Psycopg2Instrumentation(DbApi2Instrumentation):
    name = "psycopg2"

    instrument_list = [("psycopg2", "connect")]

    def call(self, module, method, wrapped, instance, args, kwargs):
        signature = "psycopg2.connect"

        host, port = get_destination_info(kwargs.get("host"), kwargs.get("port"))
        database = kwargs.get("database")
        signature = f"{signature} {host}:{port}"
        destination_info = {
            "address": host,
            "port": port,
        }
        with capture_span(
            signature,
            span_type="db",
            span_subtype="postgresql",
            span_action="connect",
            leaf=True,
            extra={"destination": destination_info, "db": {"type": "sql", "instance": database}},
        ):
            return PGConnectionProxy(wrapped(*args, **kwargs), destination_info=destination_info)


class Psycopg2ExtensionsInstrumentation(DbApi2Instrumentation):
    """
    Some extensions do a type check on the Connection/Cursor in C-code, which our
    proxy fails. For these extensions, we need to ensure that the unwrapped
    Connection/Cursor is passed.
    """

    name = "psycopg2"

    instrument_list = [
        ("psycopg2.extensions", "register_type"),
        # specifically instrument `register_json` as it bypasses `register_type`
        ("psycopg2._json", "register_json"),
        ("psycopg2.extensions", "quote_ident"),
        ("psycopg2.extensions", "encrypt_password"),
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        if "conn_or_curs" in kwargs and hasattr(kwargs["conn_or_curs"], "__wrapped__"):
            kwargs["conn_or_curs"] = kwargs["conn_or_curs"].__wrapped__
        # register_type takes the connection as second argument
        elif len(args) == 2 and hasattr(args[1], "__wrapped__"):
            args = (args[0], args[1].__wrapped__)
        # register_json takes the connection as first argument, and can have
        # several more arguments
        elif method == "register_json":
            if args and hasattr(args[0], "__wrapped__"):
                args = (args[0].__wrapped__,) + args[1:]

        elif method == "encrypt_password":
            # connection/cursor is either 3rd argument, or "scope" keyword argument
            if len(args) >= 3 and hasattr(args[2], "__wrapped__"):
                args = args[:2] + (args[2].__wrapped__,) + args[3:]
            elif "scope" in kwargs and hasattr(kwargs["scope"], "__wrapped__"):
                kwargs["scope"] = kwargs["scope"].__wrapped__

        return wrapped(*args, **kwargs)


def get_destination_info(host: Optional[str], port: Union[None, str, int]) -> tuple:
    if host:
        if "," in host:  # multiple hosts defined, take first
            host = host.split(",")[0]
    else:
        host = "localhost"
    if port:
        port = str(port)
        if "," in port:  # multiple ports defined, take first
            port = port.split(",")[0]
        port = int(port)
    else:
        port = default_ports.get("postgresql")
    return host, port



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

from atatus.contrib.asyncio.traces import async_capture_span
from atatus.instrumentation.packages.asyncio.base import AsyncAbstractInstrumentedModule


class RedisAsyncioInstrumentation(AsyncAbstractInstrumentedModule):
    name = "redis"

    instrument_list = [
        ("redis.asyncio.client", "Redis.execute_command"),
        ("redis.asyncio.client", "PubSub.execute_command"),
    ]

    async def call(self, module, method, wrapped, instance, args, kwargs):
        if len(args) > 0:
            wrapped_name = args[0]
            if isinstance(wrapped_name, bytes):
                wrapped_name = wrapped_name.decode()
        else:
            wrapped_name = self.get_wrapped_name(wrapped, instance, method)

        async with async_capture_span(
            wrapped_name, span_type="db", span_subtype="redis", span_action="query", leaf=True
        ) as span:
            if span.context is not None:
                span.context["destination"] = _get_destination_info(instance)

            return await wrapped(*args, **kwargs)


class RedisPipelineInstrumentation(AsyncAbstractInstrumentedModule):
    name = "redis"

    instrument_list = [("redis.asyncio.client", "Pipeline.execute")]

    async def call(self, module, method, wrapped, instance, args, kwargs):
        wrapped_name = self.get_wrapped_name(wrapped, instance, method)

        async with async_capture_span(
            wrapped_name, span_type="db", span_subtype="redis", span_action="query", leaf=True
        ) as span:
            if span.context is not None:
                span.context["destination"] = _get_destination_info(instance)

            return await wrapped(*args, **kwargs)


def _get_destination_info(connection):
    destination_info = {"service": {"name": "", "resource": "redis", "type": ""}}
    if connection.connection_pool:
        destination_info["port"] = connection.connection_pool.connection_kwargs.get("port")
        destination_info["address"] = connection.connection_pool.connection_kwargs.get("host")

    return destination_info




import json
import gzip
import requests
from .builder import Builder
from atatus.utils import compat
from atatus.utils.logging import get_logger
import atatus
import time

class BaseTransport(object):
    def __init__(self, config, metadata):
        self._config = config
        self._performance = self._config.performance
        self._error_logger = get_logger("atatus.errors")
        self._notify_host = config.notify_host if config.notify_host.endswith("/") else config.notify_host + "/"
        self._analytics_notify_host = config.analytics_notify_host if config.analytics_notify_host.endswith("/") else config.analytics_notify_host + "/"
        self._builder = Builder(config, metadata)
        # self._session = requests.Session()
        self._blocked = False
        self._capture_percentiles = False
        self._hostinfo_response = {}
        self._post_params = {
            'license_key': self._config.license_key,
            'agent_name': atatus.PYTHON_AGENT,
            "version": atatus.VERSION
        }

    def _get_host_info(self):
        return self._builder._common()

    def _post(self, endpoint, data):
        if endpoint != 'track/apm/analytics/txn':
            if (self._blocked is True) and (endpoint != 'track/apm/hostinfo'):
                return

        try:
            time.sleep(0.01)

            notify_host = self._notify_host
            if endpoint == 'track/apm/analytics/txn':
                notify_host = self._analytics_notify_host

            try:
                compressed_data = gzip.compress(json.dumps(data).encode("utf-8"))
                # Set the headers with 'Content-Encoding: gzip'
                headers = {
                    "Content-Encoding": "gzip",
                    "Content-Type": "application/json",
                }
                
                r = requests.post(notify_host + endpoint, params=self._post_params, timeout=30, data=compressed_data, headers=headers)
            except Exception:
                headers = {
                    "Content-Type": "application/json",
                }                
                r = requests.post(notify_host + endpoint, params=self._post_params, timeout=30, data=data, headers=headers)
                pass


            if r.status_code == 200:
                if endpoint != 'track/apm/analytics/txn':
                    self._blocked = False

                if endpoint == 'track/apm/hostinfo':
                    self._capture_percentiles = False
                    c = r.content
                    if not c:
                        return

                    c = c.decode('UTF-8')

                    resp = json.loads(c)
                    if resp:
                        if "blocked" in resp:
                            self._blocked = resp["blocked"]

                        if "tracing" in resp:
                            self._hostinfo_response["tracing"] = resp["tracing"]

                        if "analytics" in resp:
                            self._hostinfo_response["analytics"] = resp["analytics"]

                        if "capturePercentiles" in resp:
                            self._capture_percentiles = resp["capturePercentiles"]
                            self._hostinfo_response["capturePercentiles"] = self._capture_percentiles

                        if "extRequestPatterns" in resp:
                            self._hostinfo_response["extRequestPatterns"] = resp["extRequestPatterns"]

                        if "ignoreTxnNamePatterns" in resp:
                            self._hostinfo_response["ignoreTxnNamePatterns"] = resp["ignoreTxnNamePatterns"]

                        if "ignoreHTTPFailurePatterns" in resp:
                            self._hostinfo_response["ignoreHTTPFailurePatterns"] = resp["ignoreHTTPFailurePatterns"]

                        if "ignoreExceptionPatterns" in resp:
                            self._hostinfo_response["ignoreExceptionPatterns"] = resp["ignoreExceptionPatterns"]

                        if "errorLimit" in resp:
                            self._hostinfo_response["errorLimit"] = resp["errorLimit"]

                        if "performance" in resp:
                            self._hostinfo_response["performance"] = resp["performance"]

                return

            if r.status_code == 400:
                c = r.content
                if not c:
                    self._error_logger.error("Atatus transport status 400, failed without content")
                    return

                if compat.PY3:
                    c = c.decode('UTF-8')

                print_message = True
                hostinfo_analytics = False
                resp = json.loads(c)
                if resp:
                    if "tracing" in resp:
                        self._hostinfo_response["tracing"] = resp["tracing"]

                    if "analytics" in resp:
                        self._hostinfo_response["analytics"] = resp["analytics"]

                    if "blocked" in resp:
                        self._blocked = resp["blocked"]

                    if "analytics" in self._hostinfo_response:
                        hostinfo_analytics = self._hostinfo_response["analytics"]
                    
                    if self._blocked and hostinfo_analytics and (endpoint != 'track/apm/analytics/txn'):
                        print_message = False
                    
                    if print_message and "errorMessage" in resp:
                        self._error_logger.error(
                            "Atatus blocked from sending data as: %s ", resp["errorMessage"])

                if print_message:
                    self._error_logger.error("Atatus transport status 400, failed with content: %r", c)
                return

            if r.status_code != 200:
                self._error_logger.error(
                    "Atatus transport unexpected non-200 response [%s] [status_code: %r]." % (self._notify_host + endpoint, r.status_code))

        except Exception as e:
            self._error_logger.debug(
                "Atatus transport [%r] failed with exception: %r", self._notify_host + endpoint, e)
            raise

    def is_performance_disabled(self):
        return "performance" in self._hostinfo_response and self._hostinfo_response["performance"] == False

    def hostinfo(self, start_time):
        payload = self._builder.hostinfo(start_time)
        self._post('track/apm/hostinfo', payload)
        return self._hostinfo_response

    def txns(self, start_time, end_time, data):
        payload = self._builder.txns(start_time, end_time, data)
        disable_performance = self.is_performance_disabled()
        if not disable_performance and self._performance:
            self._post('track/apm/txn', payload)

    def txn_hist(self, start_time, end_time, data):
        if self._capture_percentiles is True:
            payload = self._builder.txn_hist(start_time, end_time, data)
            disable_performance = self.is_performance_disabled()
            if not disable_performance and self._performance:
                self._post('track/apm/txn/histogram', payload)

    def traces(self, start_time, end_time, data):
        payload = self._builder.traces(start_time, end_time, data)
        self._post('track/apm/trace', payload)

    def error_metrics(self, start_time, end_time, metrics_data, requests_data):
        payload = self._builder.error_metrics(start_time, end_time, metrics_data, requests_data)
        disable_performance = self.is_performance_disabled()
        if not disable_performance and self._performance:
            self._post('track/apm/error_metric', payload)

    def errors(self, start_time, end_time, error_data):
        payload = self._builder.errors(start_time, end_time, error_data)
        self._post('track/apm/error', payload)

    def metrics(self, start_time, end_time, metrics_data):
        payload = self._builder.metrics(start_time, end_time, metrics_data)
        disable_performance = self.is_performance_disabled()
        if not disable_performance and self._performance:
            self._post('track/apm/metric', payload)

    def analytics(self, start_time, end_time, analytics_data):
        payload = self._builder.analytics(start_time, end_time, analytics_data)
        disable_performance = self.is_performance_disabled()
        if not disable_performance and self._performance:
            self._post('track/apm/analytics/txn', payload)



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from atatus.instrumentation.packages.dbapi2 import (
    ConnectionProxy,
    CursorProxy,
    DbApi2Instrumentation,
    extract_signature,
)
from atatus.traces import capture_span
from atatus.utils.encoding import shorten


class SQLiteCursorProxy(CursorProxy):
    provider_name = "sqlite"

    def extract_signature(self, sql):
        return extract_signature(sql)


class SQLiteConnectionProxy(ConnectionProxy):
    cursor_proxy = SQLiteCursorProxy

    # we need to implement wrappers for the non-standard Connection.execute and
    # Connection.executemany methods

    def _trace_sql(self, method, sql, params):
        signature = extract_signature(sql)
        sql_string = shorten(sql, string_length=10000)
        with capture_span(
            signature,
            span_type="db",
            span_subtype="sqlite",
            span_action="query",
            extra={"db": {"type": "sql", "statement": sql_string}},
        ):
            if params is None:
                return method(sql)
            else:
                return method(sql, params)

    def execute(self, sql, params=None):
        return self._trace_sql(self.__wrapped__.execute, sql, params)

    def executemany(self, sql, params=None):
        return self._trace_sql(self.__wrapped__.executemany, sql, params)


class SQLiteInstrumentation(DbApi2Instrumentation):
    name = "sqlite"

    instrument_list = [("sqlite3", "connect"), ("sqlite3.dbapi2", "connect"), ("pysqlite2.dbapi2", "connect")]

    def call(self, module, method, wrapped, instance, args, kwargs):
        signature = ".".join([module, method])

        if len(args) == 1:
            signature += " " + str(args[0])

        with capture_span(signature, span_type="db", span_subtype="sqlite", span_action="connect"):
            return SQLiteConnectionProxy(wrapped(*args, **kwargs))



from importlib.metadata import Distribution, distributions
from enum import Enum
from typing import Optional, Set
# from atatus.base import get_client
from atatus.utils.logging import get_logger
from opentelemetry.semconv_ai import SpanAttributes
from opentelemetry.context import get_value, attach, set_value
from opentelemetry import trace

logging = get_logger("atatus")
def _get_package_name(dist: Distribution) -> str | None:
    try:
        return dist.name.lower()
    except (KeyError, AttributeError):
        return None

installed_packages = {name for dist in distributions() if (name := _get_package_name(dist)) is not None}


def is_package_installed(package_name: str) -> bool:
    return package_name.lower() in installed_packages


def set_association_properties(properties: dict) -> None:
    attach(set_value("association_properties", properties))

    # Attach association properties to the current span, if it's a workflow or a task
    span = trace.get_current_span()
    if get_value("workflow_name") is not None or get_value("entity_name") is not None:
        _set_association_properties_attributes(span, properties)


def _set_association_properties_attributes(span, properties: dict) -> None:
    for key, value in properties.items():
        span.set_attribute(
            f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.{key}", value
        )


def set_workflow_name(workflow_name: str) -> None:
    attach(set_value("workflow_name", workflow_name))


def set_entity_path(entity_path: str) -> None:
    attach(set_value("entity_path", entity_path))

def get_chained_entity_path(entity_name: str) -> str:
    parent = get_value("entity_path")
    if parent is None:
        return entity_name
    else:
        return f"{parent}.{entity_name}"


class Instruments(Enum):
    ALEPHALPHA = "alephalpha"
    ANTHROPIC = "anthropic"
    BEDROCK = "bedrock"
    CHROMA = "chroma"
    COHERE = "cohere"
    CREW = "crew"
    GOOGLE_GENERATIVEAI = "google_generativeai"
    GROQ = "groq"
    HAYSTACK = "haystack"
    LANCEDB = "lancedb"
    LANGCHAIN = "langchain"
    LLAMA_INDEX = "llama_index"
    MARQO = "marqo"
    MCP = "mcp"
    MILVUS = "milvus"
    MISTRAL = "mistral"
    OLLAMA = "ollama"
    OPENAI = "openai"
    PINECONE = "pinecone"
    # PYMYSQL = "pymysql"
    QDRANT = "qdrant"
    # REDIS = "redis"
    REPLICATE = "replicate"
    # REQUESTS = "requests"
    SAGEMAKER = "sagemaker"
    TOGETHER = "together"
    TRANSFORMERS = "transformers"
    # URLLIB3 = "urllib3"
    VERTEXAI = "vertexai"
    WATSONX = "watsonx"
    WEAVIATE = "weaviate"

def init_instrumentations(
    should_enrich_metrics: bool,
    instruments: Optional[Set[Instruments]] = None,
    block_instruments: Optional[Set[Instruments]] = None,
):
    block_instruments = block_instruments or set()
    instruments = instruments or set(
        Instruments
    )  # Use all instruments if none specified

    # Remove any instruments that were explicitly blocked
    instruments = instruments - block_instruments

    instrument_set = False
    for instrument in instruments:
        if instrument == Instruments.ALEPHALPHA:
            if init_alephalpha_instrumentor():
                instrument_set = True
        elif instrument == Instruments.ANTHROPIC:
            if init_anthropic_instrumentor(should_enrich_metrics):
                instrument_set = True
        elif instrument == Instruments.BEDROCK:
            if init_bedrock_instrumentor(should_enrich_metrics):
                instrument_set = True
        elif instrument == Instruments.CHROMA:
            if init_chroma_instrumentor():
                instrument_set = True
        elif instrument == Instruments.COHERE:
            if init_cohere_instrumentor():
                instrument_set = True
        elif instrument == Instruments.CREW:
            if init_crewai_instrumentor():
                instrument_set = True
        elif instrument == Instruments.GOOGLE_GENERATIVEAI:
            if init_google_generativeai_instrumentor():
                instrument_set = True
        elif instrument == Instruments.GROQ:
            if init_groq_instrumentor():
                instrument_set = True
        elif instrument == Instruments.HAYSTACK:
            if init_haystack_instrumentor():
                instrument_set = True
        elif instrument == Instruments.LANCEDB:
            if init_lancedb_instrumentor():
                instrument_set = True
        elif instrument == Instruments.LANGCHAIN:
            if init_langchain_instrumentor():
                instrument_set = True
        elif instrument == Instruments.LLAMA_INDEX:
            if init_llama_index_instrumentor():
                instrument_set = True
        elif instrument == Instruments.MARQO:
            if init_marqo_instrumentor():
                instrument_set = True
        elif instrument == Instruments.MCP:
            if init_mcp_instrumentor():
                instrument_set = True
        elif instrument == Instruments.MILVUS:
            if init_milvus_instrumentor():
                instrument_set = True
        elif instrument == Instruments.MISTRAL:
            if init_mistralai_instrumentor():
                instrument_set = True
        elif instrument == Instruments.OLLAMA:
            if init_ollama_instrumentor():
                instrument_set = True
        elif instrument == Instruments.OPENAI:
            if init_openai_instrumentor(should_enrich_metrics):
                instrument_set = True
        elif instrument == Instruments.PINECONE:
            if init_pinecone_instrumentor():
                instrument_set = True
        # elif instrument == Instruments.PYMYSQL:
        #     if init_pymysql_instrumentor():
        #         instrument_set = True
        elif instrument == Instruments.QDRANT:
            if init_qdrant_instrumentor():
                instrument_set = True
        # elif instrument == Instruments.REDIS:
        #     if init_redis_instrumentor():
        #         instrument_set = True
        elif instrument == Instruments.REPLICATE:
            if init_replicate_instrumentor():
                instrument_set = True
        # elif instrument == Instruments.REQUESTS:
        #     if init_requests_instrumentor():
        #         instrument_set = True
        elif instrument == Instruments.SAGEMAKER:
            if init_sagemaker_instrumentor(should_enrich_metrics):
                instrument_set = True
        elif instrument == Instruments.TOGETHER:
            if init_together_instrumentor():
                instrument_set = True
        elif instrument == Instruments.TRANSFORMERS:
            if init_transformers_instrumentor():
                instrument_set = True
        # elif instrument == Instruments.URLLIB3:
        #     if init_urllib3_instrumentor():
        #         instrument_set = True
        elif instrument == Instruments.VERTEXAI:
            if init_vertexai_instrumentor():
                instrument_set = True
        elif instrument == Instruments.WATSONX:
            if init_watsonx_instrumentor():
                instrument_set = True
        elif instrument == Instruments.WEAVIATE:
            if init_weaviate_instrumentor():
                instrument_set = True
        else:
            logging.warning(f"Warning: {instrument} instrumentation does not exist.")
            logging.warning(
                "Usage:\n"
                "from atatus.llmobs import Instruments\n"
                "Traceloop.init(app_name='...', instruments=set([Instruments.OPENAI]))"
            )
            # print(Fore.RESET)

    if not instrument_set:
        logging.warning(
            "Warning: No valid instruments set. ",
            "Specify instruments or remove 'instruments' argument to use all instruments."
        )
        # print(Fore.RESET)

    return instrument_set


def init_openai_instrumentor(should_enrich_metrics: bool):
    try:
        if is_package_installed("openai"):
            from opentelemetry.instrumentation.openai import OpenAIInstrumentor

            instrumentor = OpenAIInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
                enrich_assistant=should_enrich_metrics,
                get_common_metrics_attributes=metrics_common_attributes,
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True

    except Exception as e:
        logging.error(f"Error initializing OpenAI instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_anthropic_instrumentor(
    should_enrich_metrics: bool
):
    try:
        if is_package_installed("anthropic"):
            from opentelemetry.instrumentation.anthropic import AnthropicInstrumentor

            instrumentor = AnthropicInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
                get_common_metrics_attributes=metrics_common_attributes,
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Anthropic instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_cohere_instrumentor():
    try:
        if is_package_installed("cohere"):
            from opentelemetry.instrumentation.cohere import CohereInstrumentor

            instrumentor = CohereInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Cohere instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_pinecone_instrumentor():
    try:
        if is_package_installed("pinecone"):
            from opentelemetry.instrumentation.pinecone import PineconeInstrumentor

            instrumentor = PineconeInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Pinecone instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_qdrant_instrumentor():
    try:
        if is_package_installed("qdrant_client"):
            from opentelemetry.instrumentation.qdrant import QdrantInstrumentor

            instrumentor = QdrantInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
    except Exception as e:
        logging.error(f"Error initializing Qdrant instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_chroma_instrumentor():
    try:
        if is_package_installed("chromadb"):
            from opentelemetry.instrumentation.chromadb import ChromaInstrumentor

            instrumentor = ChromaInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Chroma instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_google_generativeai_instrumentor():
    try:
        if is_package_installed("google-generativeai"):
            from opentelemetry.instrumentation.google_generativeai import (
                GoogleGenerativeAiInstrumentor,
            )

            instrumentor = GoogleGenerativeAiInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Gemini instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_haystack_instrumentor():
    try:
        if is_package_installed("haystack"):
            from opentelemetry.instrumentation.haystack import HaystackInstrumentor

            instrumentor = HaystackInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Haystack instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_langchain_instrumentor():
    try:
        if is_package_installed("langchain") or is_package_installed("langgraph"):
            from opentelemetry.instrumentation.langchain import LangchainInstrumentor

            instrumentor = LangchainInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing LangChain instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_mistralai_instrumentor():
    try:
        if is_package_installed("mistralai"):
            from opentelemetry.instrumentation.mistralai import MistralAiInstrumentor

            instrumentor = MistralAiInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing MistralAI instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_ollama_instrumentor():
    try:
        if is_package_installed("ollama"):
            from opentelemetry.instrumentation.ollama import OllamaInstrumentor

            instrumentor = OllamaInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Ollama instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_transformers_instrumentor():
    try:
        if is_package_installed("transformers"):
            from opentelemetry.instrumentation.transformers import (
                TransformersInstrumentor,
            )

            instrumentor = TransformersInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Transformers instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_together_instrumentor():
    try:
        if is_package_installed("together"):
            from opentelemetry.instrumentation.together import TogetherAiInstrumentor

            instrumentor = TogetherAiInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing TogetherAI instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_llama_index_instrumentor():
    try:
        if is_package_installed("llama-index") or is_package_installed("llama_index"):
            from opentelemetry.instrumentation.llamaindex import LlamaIndexInstrumentor

            instrumentor = LlamaIndexInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing LlamaIndex instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_milvus_instrumentor():
    try:
        if is_package_installed("pymilvus"):
            from opentelemetry.instrumentation.milvus import MilvusInstrumentor

            instrumentor = MilvusInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Milvus instrumentor: {e}")
        # get_client().capture_exception()
        return False


# def init_requests_instrumentor():
#     try:
#         if is_package_installed("requests"):
#             from opentelemetry.instrumentation.requests import RequestsInstrumentor
#
#             instrumentor = RequestsInstrumentor()
#             if not instrumentor.is_instrumented_by_opentelemetry:
#                 instrumentor.instrument(excluded_urls=EXCLUDED_URLS)
#         return True
#     except Exception as e:
#         logging.error(f"Error initializing Requests instrumentor: {e}")
#         # get_client().capture_exception()
#         return False


# def init_urllib3_instrumentor():
#     try:
#         if is_package_installed("urllib3"):
#             from opentelemetry.instrumentation.urllib3 import URLLib3Instrumentor
#
#             instrumentor = URLLib3Instrumentor()
#             if not instrumentor.is_instrumented_by_opentelemetry:
#                 instrumentor.instrument(excluded_urls=EXCLUDED_URLS)
#         return True
#     except Exception as e:
#         logging.error(f"Error initializing urllib3 instrumentor: {e}")
#         # get_client().capture_exception()
#         return False
#

def init_pymysql_instrumentor():
    try:
        if is_package_installed("sqlalchemy"):
            from opentelemetry.instrumentation.sqlalchemy import SQLAlchemyInstrumentor

            instrumentor = SQLAlchemyInstrumentor()
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing SQLAlchemy instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_bedrock_instrumentor(should_enrich_metrics: bool):
    try:
        if is_package_installed("boto3"):
            from opentelemetry.instrumentation.bedrock import BedrockInstrumentor

            instrumentor = BedrockInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Bedrock instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_sagemaker_instrumentor(should_enrich_metrics: bool):
    try:
        if is_package_installed("boto3"):
            from opentelemetry.instrumentation.sagemaker import SageMakerInstrumentor

            instrumentor = SageMakerInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing SageMaker instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_replicate_instrumentor():
    try:
        if is_package_installed("replicate"):
            from opentelemetry.instrumentation.replicate import ReplicateInstrumentor

            instrumentor = ReplicateInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Replicate instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_vertexai_instrumentor():
    try:
        if is_package_installed("google-cloud-aiplatform"):
            from opentelemetry.instrumentation.vertexai import VertexAIInstrumentor

            instrumentor = VertexAIInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.warning(f"Error initializing Vertex AI instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_watsonx_instrumentor():
    try:
        if is_package_installed("ibm-watsonx-ai") or is_package_installed(
            "ibm_watson_machine_learning"
        ):
            from opentelemetry.instrumentation.watsonx import WatsonxInstrumentor

            instrumentor = WatsonxInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.warning(f"Error initializing Watsonx instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_weaviate_instrumentor():
    try:
        if is_package_installed("weaviate"):
            from opentelemetry.instrumentation.weaviate import WeaviateInstrumentor

            instrumentor = WeaviateInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.warning(f"Error initializing Weaviate instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_alephalpha_instrumentor():
    try:
        if is_package_installed("aleph_alpha_client"):
            from opentelemetry.instrumentation.alephalpha import AlephAlphaInstrumentor

            instrumentor = AlephAlphaInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Aleph Alpha instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_marqo_instrumentor():
    try:
        if is_package_installed("marqo"):
            from opentelemetry.instrumentation.marqo import MarqoInstrumentor

            instrumentor = MarqoInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing marqo instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_lancedb_instrumentor():
    try:
        if is_package_installed("lancedb"):
            from opentelemetry.instrumentation.lancedb import LanceInstrumentor

            instrumentor = LanceInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing LanceDB instrumentor: {e}")


# def init_redis_instrumentor():
#     try:
#         if is_package_installed("redis"):
#             from opentelemetry.instrumentation.redis import RedisInstrumentor
#
#             instrumentor = RedisInstrumentor()
#             if not instrumentor.is_instrumented_by_opentelemetry:
#                 instrumentor.instrument(excluded_urls=EXCLUDED_URLS)
#         return True
#     except Exception as e:
#         logging.error(f"Error initializing redis instrumentor: {e}")
#         # get_client().capture_exception()
#         return False


def init_groq_instrumentor():
    try:
        if is_package_installed("groq"):
            from opentelemetry.instrumentation.groq import GroqInstrumentor

            instrumentor = GroqInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing Groq instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_crewai_instrumentor():
    try:
        if is_package_installed("crewai"):
            from opentelemetry.instrumentation.crewai import CrewAIInstrumentor

            instrumentor = CrewAIInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing CrewAI instrumentor: {e}")
        # get_client().capture_exception()
        return False


def init_mcp_instrumentor():
    try:
        if is_package_installed("mcp"):
            from opentelemetry.instrumentation.mcp import McpInstrumentor

            instrumentor = McpInstrumentor(
               # exception_logger=lambda e: # get_client().capture_exception(),
            )
            if not instrumentor.is_instrumented_by_opentelemetry:
                instrumentor.instrument()
        return True
    except Exception as e:
        logging.error(f"Error initializing MCP instrumentor: {e}")
        # get_client().capture_exception()
        return False


def metrics_common_attributes():
    common_attributes = {}
    workflow_name = get_value("workflow_name")
    if workflow_name is not None:
        common_attributes[SpanAttributes.TRACELOOP_WORKFLOW_NAME] = workflow_name

    entity_name = get_value("entity_name")
    if entity_name is not None:
        common_attributes[SpanAttributes.TRACELOOP_ENTITY_NAME] = entity_name

    association_properties = get_value("association_properties")
    if association_properties is not None:
        for key, value in association_properties.items():
            common_attributes[
                f"{SpanAttributes.TRACELOOP_ASSOCIATION_PROPERTIES}.{key}"
            ] = value

    return common_attributes



#  BSD 3-Clause License
#
#  Copyright (c) 2021, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Tuple

from atatus.conf import constants


def get_request_data(args, kwargs) -> Tuple[Tuple[str, str, int, str], str, list]:
    if len(args) == 1 and hasattr(args[0], "method"):
        # httpcore >= 0.14
        request = args[0]
        method = request.method.decode("utf-8")
        headers = request.headers
        url = (
            request.url.scheme.decode("utf-8"),
            request.url.host.decode("utf-8"),
            request.url.port,
            request.url.target.decode("utf-8"),
        )
    else:
        if "method" in kwargs:
            method = kwargs["method"].decode("utf-8")
        else:
            method = args[0].decode("utf-8")

        # URL is a tuple of (scheme, host, port, path)
        if "url" in kwargs:
            url = kwargs["url"]
        else:
            url = args[1]
        url = (url[0].decode("utf-8"), url[1].decode("utf-8"), url[2], url[3].decode("utf-8"))
        if "headers" in kwargs:
            headers = kwargs["headers"]
        else:
            headers = []
    return url, method, headers


def get_status(response) -> int:
    if hasattr(response, "status"):  # httpcore >= 0.14
        status_code = response.status
    elif len(response) > 4:
        # httpcore < 0.11.0
        # response = (http_version, status_code, reason_phrase, headers, stream)
        status_code = response[1]
    else:
        # httpcore >= 0.11.0
        # response = (status_code, headers, stream, ext)
        status_code = response[0]
    return status_code


def set_disttracing_headers(headers, trace_parent, transaction) -> None:
    trace_parent_str = trace_parent.to_string()
    headers.append((bytes(constants.TRACEPARENT_HEADER_NAME, "utf-8"), bytes(trace_parent_str, "utf-8")))
    if transaction.tracer.config.use_atatus_traceparent_header:
        headers.append((bytes(constants.TRACEPARENT_LEGACY_HEADER_NAME, "utf-8"), bytes(trace_parent_str, "utf-8")))
    if trace_parent.tracestate:
        headers.append((bytes(constants.TRACESTATE_HEADER_NAME, "utf-8"), bytes(trace_parent.tracestate, "utf-8")))



import os
import sys
import socket
import platform
from .layer import Layer
import atatus
from atatus.utils.stacks import get_lines_from_file
from atatus.utils.package_version_utils import get_plugins

class Builder(object):
    def __init__(self, config, metadata):
        self._config = config
        self._metadata = metadata
        self._hostid = None
        if "hwinfo" in self._metadata:
            if "bootid" in self._metadata["hwinfo"]:
                self._hostid = self._metadata["hwinfo"]["bootid"]

        self._real_hostname = socket.gethostname()

        if not self._hostid:
            self._hostid = self._real_hostname

        self._containerid = None
        if "system" in self._metadata:
            if "container" in self._metadata["system"]:
                if "id" in self._metadata["system"]["container"]:
                    self._containerid = self._metadata["system"]["container"]["id"]

    def _common(self):
        common = {
            "appName": self._config.app_name,
            "licenseKey": self._config.license_key,
            "agent": {
                "name": atatus.PYTHON_AGENT,
                "version": atatus.VERSION
            },
            "hostname": self._config.hostname,
            "hostId": self._hostid,
            "version": self._config.version,
            "metadata": self._metadata
        }
        if self._containerid:
            common["containerId"] = self._containerid

        return common

    def hostinfo(self, start_time):
        payload = self._common()
        payload["timestamp"] = start_time
        payload["environment"] = self._build_hostinfo_obj()
        return payload

    def txns(self, start_time, end_time, data):
        payload = self._common()
        payload["startTime"] = start_time
        payload["endTime"] = end_time
        payload["transactions"] = self._build_txns_obj(data)
        return payload

    def txn_hist(self, start_time, end_time, data):
        payload = self._common()
        payload["startTime"] = start_time
        payload["endTime"] = end_time
        payload["transactions"] = self._build_txn_hist_obj(data)
        return payload

    def traces(self, start_time, end_time, data):
        payload = self._common()
        payload["startTime"] = start_time
        payload["endTime"] = end_time
        payload["traces"] = self._build_traces_obj(data)
        return payload

    def error_metrics(self, start_time, end_time, metrics_data, requests_data):
        payload = self._common()
        payload["startTime"] = start_time
        payload["endTime"] = end_time
        payload["errorMetrics"] = self._build_error_metrics_obj(metrics_data)
        payload["errorRequests"] = self._build_error_requests_obj(requests_data)
        return payload

    def errors(self, start_time, end_time, error_data):
        payload = self._common()
        payload["errors"] = self._build_errors_obj(error_data)
        return payload

    def metrics(self, start_time, end_time, metric_data):
        payload = self._common()
        payload["startTime"] = start_time
        payload["endTime"] = end_time
        payload["python"] = metric_data
        return payload

    def analytics(self, start_time, end_time, analytics_data):
        payload = self._common()
        payload["startTime"] = start_time
        payload["endTime"] = end_time
        payload["requests"] = analytics_data
        return payload

    def _build_hostinfo_obj(self):
        environment = {}
        environment["host"] = self._build_hostinfo_env_host_obj()
        environment["settings"] = self._build_hostinfo_settings()
        environment["python_packages"] = self._build_hostinfo_python_packages()
        return environment

    def _build_hostinfo_env_host_obj(self):
        hostinfo = {}
        hostinfo["hostname"] = self._real_hostname
        if "hwinfo" in self._metadata:
            if "cpu" in self._metadata["hwinfo"]:
                hostinfo["cpu"] = self._metadata["hwinfo"]["cpu"]
            if "mem_size" in self._metadata["hwinfo"]:
                hostinfo["ram"] = self._metadata["hwinfo"]["mem_size"]
            if "bootid" in self._metadata["hwinfo"]:
                hostinfo["bootId"] = self._metadata["hwinfo"]["bootid"]
        hostinfo["hostId"] = self._hostid
        hostinfo["arch"] = platform.machine()
        hostinfo["os"] = platform.system()
        hostinfo["kernel"] = platform.release()

        if self._containerid:
            hostinfo["containerId"] = self._containerid

        # if "system" in self._metadata:
        #     if "kubernetes" in self._metadata["system"]:
        #         hostinfo["kubernetes"] = self._metadata["system"]["kubernetes"]

        return hostinfo

    def _build_hostinfo_settings(self):
        settings = {}
        settings["pythonExecutable"] = sys.executable
        pyversion = platform.python_version_tuple()
        if pyversion:
            settings["pythonVersion"] = '.'.join(pyversion)
        settings["pythonApplication"] = sys.argv[0]
        settings["pythonImplementation"] = platform.python_implementation()

        python_home = os.environ.get("PYTHONHOME", "")
        if python_home:
            settings["envPythonHome"] = python_home

        python_path = os.environ.get("PYTHONPATH", "")
        if python_path:
            settings["envPythonPath"] = python_path

        settings["sysVersion"] = sys.version
        settings["sysPlatform"] = sys.platform
        settings["sysMaxunicode"] = sys.maxunicode

        if "service" in self._metadata:
            if "framework" in self._metadata["service"]:
                if "name" in self._metadata["service"]["framework"]:
                    settings["frameworkName"] = self._metadata["service"]["framework"]["name"]
                if "version" in self._metadata["service"]["framework"]:
                    settings["frameworkVersion"] = self._metadata["service"]["framework"]["version"]

        settings["appName"] = self._config.app_name
        settings["analytics"] = self._config.analytics
        settings["tracing"] = self._config.tracing
        settings["llmobs"] = self._config.llmobs
        settings["performance"] = self._config.performance
        settings["errorLimit"] = self._config.errorLimit
        settings["logBody"] = self._config.capture_body
        settings["traceThreshold"] = (self._config.trace_threshold.total_seconds() * 1000)
        settings["agentVersion"] = atatus.VERSION

        return settings

    def _build_hostinfo_python_packages(self):
        try:
            return get_plugins()
        except Exception as e:
            self._error_logger.debug("An error occurred while getting plugins: %r" % e)
            return {}

    def _build_metric(self, name, value, background=False):
        if not name or not value:
            return {}

        metric = {
            "name": name,
            "type": value.type,
            "kind": value.kind,
            "durations": [value.count, value.total, value.min, value.max]
        }

        if background == True:
            metric["background"] = background

        return metric

    def _build_hist_metric(self, name, value):
        if not name or not value:
            return {}

        hist_metric = {
            "name": name,
            "type": value.type,
            "kind": value.kind,
            "histogram": value.hist
        }

        return hist_metric

    def _build_request(self, context):
        request = {}
        if "request" in context:
            r = context["request"]
            request["method"] = r.get("method", "")
            if "headers" in r:
                h = r["headers"]
                request["accept"] = h.get("accept", "")
                request["accept-encoding"] = h.get("accept-encoding", "")
                request["accept-language"] = h.get("accept-language", "")
                request["userAgent"] = h.get("user-agent", "")

            if "url" in r:
                u = r["url"]
                request["host"] = u.get("hostname", "")
                request["port"] = int(u.get("port", 0))
                request["path"] = u.get("pathname", "")

        if "response" in context:
            if "status_code" in context["response"]:
                request["statusCode"] = context["response"]["status_code"]

        return request

    def _build_custom_data(self, context):
        custom_data = {}
        if "custom" in context and isinstance(context["custom"], dict):
            for key, value in context["custom"].items():
                try:
                    custom_data[str(key)] = str(value)
                except Exception as e:
                    continue

        return custom_data

    def _build_user(self, context):
        user = {}
        if "user" in context:
            if "id" in context["user"]:
                user["id"] = str(context["user"]["id"])

            if "username" in context["user"]:
                user["username"] = context["user"]["username"]

            if "email" in context["user"]:
                user["email"] = context["user"]["email"]

        return user

    def _build_txns_obj(self, data):
        txns = []
        for t, v in data.items():
            txn = self._build_metric(t, v, v.background)
            txn["traces"] = []
            for l, u in v.spans.items():
                txn["traces"].append(self._build_metric(l, u))
            txns.append(txn)
        return txns

    def _build_txn_hist_obj(self, data):
        txn_hists = []
        for t, v in data.items():
            hist = self._build_hist_metric(t, v)
            if hist:
                txn_hists.append(hist)

        return txn_hists

    def _build_traces_obj(self, traces_data):
        traces = []

        for txn in traces_data:
            if not all(k in txn for k in ("name", "id", "timestamp", "duration")):
                continue

            trace = {}
            trace["name"] = txn["name"]
            trace["type"] = self._config.framework_name or atatus.PYTHON_AGENT
            trace["kind"] = atatus.PYTHON_AGENT
            if "background" in txn:
                trace["background"] = txn["background"]
            trace["start"] = (txn["timestamp"] // 1000)
            trace["duration"] = txn["duration"]
            if "context" in txn:
                trace["request"] = self._build_request(txn["context"])
                trace["customData"] = self._build_custom_data(txn["context"])
                trace["user"] = self._build_user(txn["context"])

            trace["txnId"] = txn["id"]
            if "trace_id" in txn:
                trace["traceId"] = txn["trace_id"]
            trace["entries"] = []
            trace["funcs"] = []

            i = 0

            if "spans" in txn:
                for span in txn["spans"]:
                    if not all(k in span for k in ("timestamp", "duration", "name", "subtype", "type")):
                        continue

                    entry = {}
                    entry["lv"] = 1
                    if span["timestamp"] >= txn["timestamp"]:
                        entry["so"] = ((span["timestamp"] - txn["timestamp"]) / 1000)
                    else:
                        entry["so"] = 0
                    entry["du"] = span["duration"]
                    entry["ly"] = {}
                    entry["ly"]["name"] = span["name"]
                    entry["ly"]["type"] = Layer.types_dict.get(span["subtype"], span["subtype"])
                    entry["ly"]["kind"] = Layer.kinds_dict.get(span["type"], span["type"])
                    if "context" in span:
                        if "db" in span["context"]:
                            if "dt" not in entry:
                                entry["dt"] = {}
                            entry["dt"]["db"] = {}
                            if "statement" in span["context"]["db"]:
                                if span["context"]["db"]["statement"] is not None:
                                    entry["dt"]["db"]["statement"] = span["context"]["db"]["statement"]
                            if "type" in span["context"]["db"]:
                                if span["context"]["db"]["type"] is not None:
                                    entry["dt"]["db"]["type"] = span["context"]["db"]["type"]
                            if "instance" in span["context"]["db"]:
                                if span["context"]["db"]["instance"] is not None:
                                    entry["dt"]["db"]["instance"] = span["context"]["db"]["instance"]
                            if "user" in span["context"]["db"]:
                                if span["context"]["db"]["user"] is not None:
                                    entry["dt"]["db"]["user"] = span["context"]["db"]["user"]
                            if "link" in span["context"]["db"]:
                                if span["context"]["db"]["link"] is not None:
                                    entry["dt"]["db"]["link"] = span["context"]["db"]["link"]
                            if "rows_affected" in span["context"]["db"]:
                                if span["context"]["db"]["rows_affected"] is not None:
                                    entry["dt"]["db"]["rows_affected"] = span["context"]["db"]["rows_affected"]

                        if "http" in span["context"]:
                            if "dt" not in entry:
                                entry["dt"] = {}
                            entry["dt"]["http"] = {}
                            if "url" in span["context"]["http"]:
                                if span["context"]["http"]["url"] is not None:
                                    entry["dt"]["http"]["url"] = span["context"]["http"]["url"]
                            if "status_code" in span["context"]["http"]:
                                if span["context"]["http"]["status_code"] is not None:
                                    entry["dt"]["http"]["status_code"] = span["context"]["http"]["status_code"]

                        if "destination" in span["context"]:
                            if "address" in span["context"]["destination"]:
                                if span["context"]["destination"]["address"] is not None:
                                    if "dt" not in entry:
                                        entry["dt"] = {}
                                    if "destination" not in entry["dt"]:
                                        entry["dt"]["destination"] = {}
                                    entry["dt"]["destination"]["address"] = span["context"]["destination"]["address"]
                            if "port" in span["context"]["destination"]:
                                if span["context"]["destination"]["port"] is not None:
                                    if "dt" not in entry:
                                        entry["dt"] = {}
                                    if "destination" not in entry["dt"]:
                                        entry["dt"]["destination"] = {}
                                    entry["dt"]["destination"]["port"] = span["context"]["destination"]["port"]

                    try:
                        func_index = trace["funcs"].index(span["name"])
                    except ValueError:
                        trace["funcs"].append(span["name"])
                        func_index = i
                        i = i + 1
                    entry["i"] = func_index
                    trace["entries"].append(entry)

            if "python_time" in txn:
                entry = {}
                entry["lv"] = 1
                entry["so"] = 0
                entry["du"] = txn["python_time"]
                entry["ly"] = {}
                entry["ly"]["name"] = atatus.PYTHON_AGENT
                entry["ly"]["type"] = atatus.PYTHON_AGENT
                entry["ly"]["kind"] = atatus.PYTHON_AGENT
                try:
                    func_index = trace["funcs"].index(atatus.PYTHON_AGENT)
                except ValueError:
                    trace["funcs"].append(atatus.PYTHON_AGENT)
                    func_index = i
                    i = i + 1
                entry["i"] = func_index
                trace["entries"].append(entry)

            traces.append(trace)

        return traces

    def _build_error_metrics_obj(self, metrics_data):
        error_metrics = []

        for t, v in metrics_data.items():
            error_metric = {}
            error_metric["name"] = t
            error_metric["type"] = self._config.framework_name or atatus.PYTHON_AGENT
            error_metric["kind"] = atatus.PYTHON_AGENT
            error_metric["statusCodes"] = v
            error_metrics.append(error_metric)

        return error_metrics

    def _build_error_requests_obj(self, requests_data):
        error_requests = []

        for v in requests_data:
            if not all(k in v for k in ("name", "timestamp", "context")):
                continue

            error_request = {}
            error_request["name"] = v["name"]
            error_request["type"] = self._config.framework_name or atatus.PYTHON_AGENT
            error_request["kind"] = atatus.PYTHON_AGENT
            error_request["timestamp"] = (v["timestamp"] // 1000)
            error_request["request"] = self._build_request(v["context"])
            error_request["customData"]  = self._build_custom_data(v["context"])
            error_request["user"]  = self._build_user(v["context"])

            error_requests.append(error_request)

        return error_requests

    def _build_errors_obj(self, errors_data):
        errors = []
        for v in errors_data:
            if "timestamp" not in v:
                continue

            errorObj = {}
            if "exception" in v:
                errorObj = v["exception"]
            elif "log" in v:
                errorObj = v["log"]
            else:
                continue

            if "message" not in errorObj:
                continue

            error = {}
            error["timestamp"] = (v["timestamp"] // 1000)

            if "event_transaction_name" in v:
                error["transaction"] = v["event_transaction_name"]

            if "background" in v:
                error["background"] = v["background"]
            if "transaction_id" in v:
                error["txnId"] = v["transaction_id"]

            if "span_id" in v:
                error["spanId"] = v["span_id"]
            if "trace_id" in v:
                error["traceId"] = v["trace_id"]

            if "context" in v:
                error["request"] = self._build_request(v["context"])
                error["customData"]  = self._build_custom_data(v["context"])
                error["user"]  = self._build_user(v["context"])

            error["exceptions"] = []
            exception = {}
            if "type" in errorObj:
                exception["class"] = errorObj["type"]
            elif "level" in errorObj:
                exception["class"] = errorObj["level"]
            else:
                exception["class"] = "Unknown"

            exception["message"] = errorObj["message"]
            exception["stacktrace"] = []
            if "stacktrace" in errorObj:
                for f in errorObj["stacktrace"]:
                    if not all(k in f for k in ("filename", "function", "lineno", "library_frame")):
                        continue

                    frame = {}
                    frame["f"] = f["filename"]
                    frame["m"] = f["function"]
                    frame["ln"] = f["lineno"]
                    if f["library_frame"] is False:
                        frame["inp"] = True

                    if "context_metadata" in f:
                        fname, lineno, context_lines, loader, module_name = f.pop("context_metadata")
                        pre_context, context_line, post_context = get_lines_from_file(
                            fname, lineno, context_lines, loader, module_name
                        )
                        if context_line:
                            f["pre_context"] = pre_context
                            f["context_line"] = context_line
                            f["post_context"] = post_context

                        if "context_line" in f:
                            frame["code"] = []

                            if "pre_context" in f:
                                psize = len(f["pre_context"])
                                lineno = 0
                                if f["lineno"] - psize > 0:
                                    lineno = f["lineno"] - psize
                                for c in f["pre_context"]:
                                    frame["code"].append([str(lineno), c])
                                    lineno += 1

                            frame["code"].append([str(f["lineno"]), f["context_line"]])

                            if "post_context" in f:
                                psize = len(f["post_context"])
                                lineno = f["lineno"] + 1
                                for c in f["post_context"]:
                                    frame["code"].append([str(lineno), c])
                                    lineno += 1

                    exception["stacktrace"].append(frame)

            error["exceptions"].append(exception)
            errors.append(error)
        return errors



#  BSD 3-Clause License
#
#  Copyright (c) 2012, the Sentry Team, see AUTHORS for more details
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE

import typing as t

from sanic import Sanic
from sanic.blueprint_group import BlueprintGroup
from sanic.blueprints import Blueprint
from sanic.request import Request
from sanic.response import HTTPResponse

UserInfoType = t.Tuple[t.Optional[t.Any], t.Optional[t.Any], t.Optional[t.Any]]
LabelInfoType = t.Dict[str, t.Union[str, bool, int, float]]
CustomInfoType = t.Dict[str, t.Any]

SanicRequestOrResponse = t.Union[Request, HTTPResponse]

ApmHandlerType = t.Optional[t.Callable[[Request, BaseException], t.Coroutine[t.Any, t.Any, None]]]

EnvInfoType = t.Iterable[t.Tuple[str, str]]

TransactionNameCallbackType = t.Optional[t.Callable[[Request], str]]

UserInfoCallbackType = t.Optional[t.Callable[[Request], t.Awaitable[UserInfoType]]]

CustomContextCallbackType = t.Optional[t.Callable[[SanicRequestOrResponse], t.Awaitable[CustomInfoType]]]

LabelInfoCallbackType = t.Optional[t.Callable[[SanicRequestOrResponse], t.Awaitable[LabelInfoType]]]

APMConfigType = t.Optional[t.Union[t.Dict[str, t.Any], t.Dict[bytes, t.Any]]]

ExtendableMiddlewareGroup = t.Union[Blueprint, BlueprintGroup]

AllMiddlewareGroup = t.Union[Sanic, Blueprint, BlueprintGroup]



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from opentracing.span import Span as OTSpanBase
from opentracing.span import SpanContext as OTSpanContextBase

from atatus import traces
from atatus.utils import get_url_dict
from atatus.utils.logging import get_logger

try:
    # opentracing-python 2.1+
    from opentracing import logs as ot_logs
    from opentracing import tags
except ImportError:
    # opentracing-python <2.1
    from opentracing.ext import tags

    ot_logs = None


logger = get_logger("atatus.contrib.opentracing")


class OTSpan(OTSpanBase):
    def __init__(self, tracer, context, atatus_ref) -> None:
        super(OTSpan, self).__init__(tracer, context)
        self.atatus_ref = atatus_ref
        self.is_transaction = isinstance(atatus_ref, traces.Transaction)
        self.is_dropped = isinstance(atatus_ref, traces.DroppedSpan)
        if not context.span:
            context.span = self

    def log_kv(self, key_values, timestamp=None):
        exc_type, exc_val, exc_tb = None, None, None
        if "python.exception.type" in key_values:
            exc_type = key_values["python.exception.type"]
            exc_val = key_values.get("python.exception.val")
            exc_tb = key_values.get("python.exception.tb")
        elif ot_logs and key_values.get(ot_logs.EVENT) == tags.ERROR:
            exc_type = key_values[ot_logs.ERROR_KIND]
            exc_val = key_values.get(ot_logs.ERROR_OBJECT)
            exc_tb = key_values.get(ot_logs.STACK)
        else:
            logger.debug("Can't handle non-exception type opentracing logs")
        if exc_type:
            agent = self.tracer._agent
            agent.capture_exception(exc_info=(exc_type, exc_val, exc_tb))
        return self

    def set_operation_name(self, operation_name):
        self.atatus_ref.name = operation_name
        return self

    def set_tag(self, key, value):
        if self.is_transaction:
            if key == "type":
                self.atatus_ref.transaction_type = value
            elif key == "result":
                self.atatus_ref.result = value
            elif key == tags.HTTP_STATUS_CODE:
                self.atatus_ref.result = "HTTP {}xx".format(str(value)[0])
                traces.set_context({"status_code": value}, "response")
            elif key == "user.id":
                traces.set_user_context(user_id=value)
            elif key == "user.username":
                traces.set_user_context(username=value)
            elif key == "user.email":
                traces.set_user_context(email=value)
            elif key == tags.HTTP_URL:
                traces.set_context({"url": get_url_dict(value)}, "request")
            elif key == tags.HTTP_METHOD:
                traces.set_context({"method": value}, "request")
            elif key == tags.COMPONENT:
                traces.set_context({"framework": {"name": value}}, "service")
            else:
                self.atatus_ref.label(**{key: value})
        elif not self.is_dropped:
            if key.startswith("db."):
                span_context = self.atatus_ref.context or {}
                if "db" not in span_context:
                    span_context["db"] = {}
                if key == tags.DATABASE_STATEMENT:
                    span_context["db"]["statement"] = value
                elif key == tags.DATABASE_USER:
                    span_context["db"]["user"] = value
                elif key == tags.DATABASE_TYPE:
                    span_context["db"]["type"] = value
                    self.atatus_ref.type = "db." + value
                else:
                    self.atatus_ref.label(**{key: value})
                self.atatus_ref.context = span_context
            elif key == tags.SPAN_KIND:
                self.atatus_ref.type = value
            else:
                self.atatus_ref.label(**{key: value})
        return self

    def finish(self, finish_time=None) -> None:
        if self.is_transaction:
            self.tracer._agent.end_transaction()
        elif not self.is_dropped:
            self.atatus_ref.transaction.end_span()


class OTSpanContext(OTSpanContextBase):
    def __init__(self, trace_parent, span=None) -> None:
        self.trace_parent = trace_parent
        self.span = span



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from atatus.metrics.base_metrics import MetricSet

try:
    import psutil
except ImportError:
    raise ImportError("psutil not found. Install it to get system and process metrics")


class CPUMetricSet(MetricSet):
    def __init__(self, registry) -> None:
        psutil.cpu_percent(interval=None)
        self._process = psutil.Process()
        self._process.cpu_percent(interval=None)
        super(CPUMetricSet, self).__init__(registry)

    def before_collect(self) -> None:
        self.gauge("system.cpu.total.norm.pct").val = psutil.cpu_percent(interval=None) / 100.0
        self.gauge("system.memory.actual.free").val = psutil.virtual_memory().available
        self.gauge("system.memory.total").val = psutil.virtual_memory().total
        p = self._process
        if hasattr(p, "oneshot"):  # new in psutil 5.0
            with p.oneshot():
                memory_info = p.memory_info()
                cpu_percent = p.cpu_percent(interval=None)
        else:
            memory_info = p.memory_info()
            cpu_percent = p.cpu_percent(interval=None)
        self.gauge("system.process.cpu.total.norm.pct").val = cpu_percent / 100.0 / psutil.cpu_count()
        self.gauge("system.process.memory.size").val = memory_info.vms
        self.gauge("system.process.memory.rss.bytes").val = memory_info.rss



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import typing
import uuid

from opentelemetry.context import Context
from opentelemetry.trace.propagation import _SPAN_KEY

from atatus.contrib.opentelemetry.span import Span as OtelSpan
from atatus.traces import Transaction, execution_context

logger = logging.getLogger("atatus.otel")


def create_key(keyname: str) -> str:
    """To allow cross-cutting concern to control access to their local state,
    the RuntimeContext API provides a function which takes a keyname as input,
    and returns a unique key.
    Args:
        keyname: The key name is for debugging purposes and is not required to be unique.
    Returns:
        A unique string representing the newly created key.
    """
    return keyname + "-" + str(uuid.uuid4())


def get_value(key: str, context: typing.Optional[Context] = None) -> "object":
    """To access the local state of a concern, the RuntimeContext API
    provides a function which takes a context and a key as input,
    and returns a value.
    Args:
        key: The key of the value to retrieve.
        context: The context from which to retrieve the value, if None, the current context is used.
    Returns:
        The value associated with the key.
    """
    return context.get(key) if context is not None else get_current().get(key)


def set_value(key: str, value: "object", context: typing.Optional[Context] = None) -> Context:
    """To record the local state of a cross-cutting concern, the
    RuntimeContext API provides a function which takes a context, a
    key, and a value as input, and returns an updated context
    which contains the new value.
    Args:
        key: The key of the entry to set.
        value: The value of the entry to set.
        context: The context to copy, if None, the current context is used.
    Returns:
        A new `Context` containing the value set.
    """
    if context is None:
        context = get_current()
    new_values = context.copy()
    new_values[key] = value
    return Context(new_values)


def get_current() -> Context:
    """To access the context associated with program execution,
    the Context API provides a function which takes no arguments
    and returns a Context.

    Returns:
        The current `Context` object.
    """
    span = execution_context.get_span()
    if not span:
        span = execution_context.get_transaction()
    if not span:
        return Context()

    otel_span = getattr(span, "otel_wrapper", OtelSpan(span.name, span))
    context = otel_span.otel_context

    return context


def attach(context: Context) -> object:
    """Associates a Context with the caller's current execution unit.

    Due to limitations in the Atatus context management, a token is not
    returned by this method, nor required to detach() a Context later.

    Note that a Context will not be attached if it doesn't have an OtelSpan at _SPAN_KEY

    Args:
        context: The Context to set as current.
    Returns:
        None
    """
    span = context.get(_SPAN_KEY)
    if not span:
        logger.error("Attempted to attach a context without a valid OtelSpan")
        return None
    span.otel_context = context
    atatus_span = span.atatus_span
    if isinstance(atatus_span, Transaction):
        execution_context.set_transaction(atatus_span)
    else:
        execution_context.set_span(atatus_span)

    return None


def detach(token: typing.Optional[object] = None) -> None:
    """Resets the Context associated with the caller's current execution unit
    to the value it had before attaching a specified Context.

    Due to limitations in the Atatus context management, a token is not
    returned by attach(), nor required to detach() a Context later.

    Args:
        token: Tokens are not supported in this bridge, this argument is unused
    """
    if execution_context.get_span():
        execution_context.unset_span()
    else:
        logger.warning("Can't detach a running transaction. Please end the transaction instead.")



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from atatus.instrumentation.packages.base import AbstractInstrumentedModule
from atatus.traces import capture_span


class PyMemcacheInstrumentation(AbstractInstrumentedModule):
    name = "pymemcache"

    method_list = [
        "add",
        "append",
        "cas",
        "decr",
        "delete",
        "delete_many",
        "delete_multi",
        "flush_all",
        "get",
        "get_many",
        "get_multi",
        "gets",
        "gets_many",
        "incr",
        "prepend",
        "quit",
        "replace",
        "set",
        "set_many",
        "set_multi",
        "stats",
        "touch",
    ]

    def get_instrument_list(self):
        return (
            [("pymemcache.client.base", "Client." + method) for method in self.method_list]
            + [("pymemcache.client.base", "PooledClient." + method) for method in self.method_list]
            + [("pymemcache.client.hash", "HashClient." + method) for method in self.method_list]
        )

    def call(self, module, method, wrapped, instance, args, kwargs):
        name = self.get_wrapped_name(wrapped, instance, method)

        # Since HashClient uses Client/PooledClient for the actual calls, we
        # don't need to get address/port info for that class
        address, port = None, None
        if getattr(instance, "server", None):
            if isinstance(instance.server, (list, tuple)):
                # Address/port are a tuple
                address, port = instance.server
            else:
                # Server is a UNIX domain socket
                address = instance.server
        destination = {
            "address": address,
            "port": port,
        }

        if "PooledClient" in name:
            # PooledClient calls out to Client for the "work", but only once,
            # so we don't care about the "duplicate" spans from Client in that
            # case
            with capture_span(
                name,
                span_type="cache",
                span_subtype="memcached",
                span_action="query",
                extra={"destination": destination},
                leaf=True,
            ):
                return wrapped(*args, **kwargs)
        else:
            with capture_span(
                name,
                span_type="cache",
                span_subtype="memcached",
                span_action="query",
                extra={"destination": destination},
                leaf=True,
            ):
                return wrapped(*args, **kwargs)



# Copyright 2010 New Relic, Inc.
# Copyright 2025 Atatus.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import types
import warnings
from atatus.packages.isort import stdlibs as isort_stdlibs

try:
    from functools import cache as _cache_package_versions
except ImportError:
    from functools import wraps
    from threading import Lock

    _package_version_cache = {}
    _package_version_cache_lock = Lock()

    def _cache_package_versions(wrapped):
        """
        Threadsafe implementation of caching for _get_package_version.

        Python 2.7 does not have the @functools.cache decorator, and
        must be reimplemented with support for clearing the cache.
        """

        @wraps(wrapped)
        def _wrapper(name):
            if name in _package_version_cache:
                return _package_version_cache[name]

            with _package_version_cache_lock:
                if name in _package_version_cache:
                    return _package_version_cache[name]

                version = _package_version_cache[name] = wrapped(name)
                return version

        def cache_clear():
            """Cache clear function to mimic @functools.cache"""
            with _package_version_cache_lock:
                _package_version_cache.clear()

        _wrapper.cache_clear = cache_clear
        return _wrapper


# Need to account for 4 possible variations of version declaration specified in (rejected) PEP 396
VERSION_ATTRS = ("__version__", "version", "__version_tuple__", "version_tuple")  # nosec
NULL_VERSIONS = frozenset((None, "", "0", "0.0", "0.0.0", "0.0.0.0", (0,), (0, 0), (0, 0, 0), (0, 0, 0, 0)))  # nosec


def get_package_version(name):
    """Gets the version string of the library.
    :param name: The name of library.
    :type name: str
    :return: The version of the library. Returns None if can't determine version.
    :type return: str or None

    Usage::
        >>> get_package_version("botocore")
                "1.1.0"
    """

    version = _get_package_version(name)

    # Coerce iterables into a string
    if isinstance(version, tuple):
        version = ".".join(str(v) for v in version)

    return version


def get_package_version_tuple(name):
    """Gets the version tuple of the library.
    :param name: The name of library.
    :type name: str
    :return: The version of the library. Returns None if can't determine version.
    :type return: tuple or None

    Usage::
        >>> get_package_version_tuple("botocore")
                (1, 1, 0)
    """

    def int_or_str(value):
        try:
            return int(value)
        except Exception:
            return str(value)

    version = _get_package_version(name)

    # Split "." separated strings and cast fields to ints
    if isinstance(version, str):
        version = tuple(int_or_str(v) for v in version.split("."))

    return version


@_cache_package_versions
def _get_package_version(name):
    module = sys.modules.get(name, None)
    version = None

    # importlib was introduced into the standard library starting in Python3.8.
    if "importlib" in sys.modules and hasattr(sys.modules["importlib"], "metadata"):
        try:
            # In Python3.10+ packages_distribution can be checked for as well
            if hasattr(sys.modules["importlib"].metadata, "packages_distributions"):  # pylint: disable=E1101
                distributions = sys.modules["importlib"].metadata.packages_distributions()  # pylint: disable=E1101
                distribution_name = distributions.get(name, name)
                distribution_name = distribution_name[0] if isinstance(distribution_name, list) else distribution_name
            else:
                distribution_name = name

            version = sys.modules["importlib"].metadata.version(distribution_name)  # pylint: disable=E1101
            if version not in NULL_VERSIONS:
                return version
        except Exception:
            pass

    with warnings.catch_warnings(record=True):
        for attr in VERSION_ATTRS:
            try:
                version = getattr(module, attr, None)

                # In certain cases like importlib_metadata.version, version is a callable
                # function.
                if callable(version):
                    continue

                # In certain cases like clickhouse-connect python library, we get the version
                # as a module instead of its version, so ignoring it.
                if isinstance(version, types.ModuleType):
                    continue

                # Cast any version specified as a list into a tuple.
                version = tuple(version) if isinstance(version, list) else version
                if version not in NULL_VERSIONS:
                    return version
            except Exception:
                pass

    if "pkg_resources" in sys.modules:
        try:
            version = sys.modules["pkg_resources"].get_distribution(name).version
            if version not in NULL_VERSIONS:
                return version
        except Exception:
            pass

def get_plugins():
    stdlib_builtin_module_names = _get_stdlib_builtin_module_names()

    plugins = {}

    # Using any iterable to create a snapshot of sys.modules can occassionally
    # fail in a rare case when modules are imported in parallel by different
    # threads.
    #
    # TL;DR: Do NOT use an iterable on the original sys.modules to generate the
    # list
    for name, module in sys.modules.copy().items():
        if "." in name or name.startswith("_"):
            continue

        # If the module isn't actually loaded (such as failed relative imports
        # in Python 2.7), the module will be None and should not be reported.
        try:
            if not module:
                continue
        except Exception:
            # if the application uses generalimport to manage optional depedencies,
            # it's possible that generalimport.MissingOptionalDependency is raised.
            # In this case, we should not report the module as it is not actually loaded and
            # is not a runtime dependency of the application.
            #
            continue

        # Exclude standard library/built-in modules.
        if name in stdlib_builtin_module_names:
            continue

        version = None
        # Attempt to look up version information
        try:
            version = get_package_version(name)
        except Exception:
            pass

        # If it has no version it's likely not a real package so don't report it
        if version:
            plugins[name] = version

    return plugins

def _get_stdlib_builtin_module_names():
    builtins = set(sys.builtin_module_names)
    # Since sys.stdlib_module_names is not available in versions of python below 3.10,
    # use isort's hardcoded stdlibs instead.
    python_version = sys.version_info[0:2]
    if python_version < (3,):
        stdlibs = isort_stdlibs.py27.stdlib
    elif (3, 7) <= python_version < (3, 8):
        stdlibs = isort_stdlibs.py37.stdlib
    elif python_version < (3, 9):
        stdlibs = isort_stdlibs.py38.stdlib
    elif python_version < (3, 10):
        stdlibs = isort_stdlibs.py39.stdlib
    elif python_version >= (3, 10):
        stdlibs = sys.stdlib_module_names
    else:
        _logger.warn("Unsupported Python version. Unable to determine stdlibs.")
        return builtins
    return builtins | stdlibs


#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Provides classes to instrument dbapi2 providers

https://www.python.org/dev/peps/pep-0249/
"""

import re
import string

import wrapt

from atatus.instrumentation.packages.base import AbstractInstrumentedModule
from atatus.traces import capture_span
from atatus.utils.encoding import force_text, shorten


class Literal(object):
    def __init__(self, literal_type, content) -> None:
        self.literal_type = literal_type
        self.content = content

    def __eq__(self, other):
        return isinstance(other, Literal) and self.literal_type == other.literal_type and self.content == other.content

    def __repr__(self):
        return "<Literal {}{}{}>".format(self.literal_type, self.content, self.literal_type)


def look_for_table(sql, keyword):
    tokens = tokenize(sql)
    table_name = _scan_for_table_with_tokens(tokens, keyword)
    if isinstance(table_name, Literal):
        table_name = table_name.content.strip(table_name.literal_type)
    return table_name


def _scan_for_table_with_tokens(tokens, keyword):
    seen_keyword = False
    for idx, lexeme in scan(tokens):
        if seen_keyword:
            if lexeme == "(":
                return _scan_for_table_with_tokens(tokens[idx:], keyword)
            else:
                return lexeme

        if isinstance(lexeme, str) and lexeme.upper() == keyword:
            seen_keyword = True


def tokenize(sql):
    # split on anything that is not a word character or a square bracket, excluding dots
    return [t for t in re.split(r"([^\w.\[\]])", sql) if t != ""]


def scan(tokens):
    literal_start_idx = None
    literal_started = None
    prev_was_escape = False
    lexeme = []
    digits = set(string.digits)

    i = 0
    while i < len(tokens):
        token = tokens[i]
        if literal_start_idx:
            if prev_was_escape:
                prev_was_escape = False
                lexeme.append(token)
            else:
                if token == literal_started:
                    if literal_started == "'" and len(tokens) > i + 1 and tokens[i + 1] == "'":  # double quotes
                        i += 1
                        lexeme.append("'")
                    else:
                        yield i, Literal(literal_started, "".join(lexeme))
                        literal_start_idx = None
                        literal_started = None
                        lexeme = []
                else:
                    if token == "\\":
                        prev_was_escape = token
                    else:
                        prev_was_escape = False
                        lexeme.append(token)
        elif literal_start_idx is None:
            if token in ["'", '"', "`"]:
                literal_start_idx = i
                literal_started = token
            elif token == "$":
                # exclude query parameters that have a digit following the dollar
                if True and len(tokens) > i + 1 and tokens[i + 1] in digits:
                    yield i, token
                    i += 1
                    continue
                # Postgres can use arbitrary characters between two $'s as a
                # literal separation token, e.g.: $fish$ literal $fish$
                # This part will detect that and skip over the literal.
                try:
                    # Closing dollar of the opening quote,
                    # i.e. the second $ in the first $fish$
                    closing_dollar_idx = tokens.index("$", i + 1)
                except ValueError:
                    pass
                else:
                    quote = tokens[i : closing_dollar_idx + 1]
                    length = len(quote)
                    # Opening dollar of the closing quote,
                    # i.e. the first $ in the second $fish$
                    closing_quote_idx = closing_dollar_idx + 1
                    while True:
                        try:
                            closing_quote_idx = tokens.index("$", closing_quote_idx)
                        except ValueError:
                            break
                        if tokens[closing_quote_idx : closing_quote_idx + length] == quote:
                            yield i, Literal(
                                "".join(quote), "".join(tokens[closing_dollar_idx + 1 : closing_quote_idx])
                            )
                            i = closing_quote_idx + length
                            break
                        closing_quote_idx += 1
            else:
                if token != " ":
                    yield i, token
        i += 1

    if lexeme:
        yield i, lexeme


def extract_signature(sql):
    """
    Extracts a minimal signature from a given SQL query
    :param sql: the SQL statement
    :return: a string representing the signature
    """
    sql = force_text(sql)
    sql = sql.strip()
    first_space = sql.find(" ")
    if first_space < 0:
        return sql

    second_space = sql.find(" ", first_space + 1)

    sql_type = sql[0:first_space].upper()

    if sql_type in ["INSERT", "DELETE"]:
        keyword = "INTO" if sql_type == "INSERT" else "FROM"
        sql_type = sql_type + " " + keyword

        object_name = look_for_table(sql, keyword)
    elif sql_type in ["CREATE", "DROP"]:
        # 2nd word is part of SQL type
        sql_type = sql_type + sql[first_space:second_space]
        object_name = ""
    elif sql_type == "UPDATE":
        object_name = look_for_table(sql, "UPDATE")
    elif sql_type == "SELECT":
        # Name is first table
        try:
            sql_type = "SELECT FROM"
            object_name = look_for_table(sql, "FROM")
        except Exception:
            object_name = ""
    elif sql_type in ["EXEC", "EXECUTE"]:
        sql_type = "EXECUTE"
        end = second_space if second_space > first_space else len(sql)
        object_name = sql[first_space + 1 : end]
    elif sql_type == "CALL":
        first_paren = sql.find("(", first_space)
        end = first_paren if first_paren > first_space else len(sql)
        procedure_name = sql[first_space + 1 : end].rstrip(";")
        object_name = procedure_name + "()"
    else:
        # No name
        object_name = ""

    signature = " ".join(filter(bool, [sql_type, object_name]))
    return signature


QUERY_ACTION = "query"
EXEC_ACTION = "exec"
PROCEDURE_STATEMENTS = ["EXEC", "EXECUTE", "CALL"]


def extract_action_from_signature(signature, default):
    if signature.split(" ")[0] in PROCEDURE_STATEMENTS:
        return EXEC_ACTION
    return default


class CursorProxy(wrapt.ObjectProxy):
    provider_name = None
    DML_QUERIES = ("INSERT", "DELETE", "UPDATE")

    def __init__(self, wrapped, destination_info=None) -> None:
        super(CursorProxy, self).__init__(wrapped)
        self._self_destination_info = destination_info or {}

    def callproc(self, procname, params=None):
        return self._trace_sql(self.__wrapped__.callproc, procname, params, action=EXEC_ACTION)

    def execute(self, sql, params=None):
        return self._trace_sql(self.__wrapped__.execute, sql, params)

    def executemany(self, sql, param_list):
        return self._trace_sql(self.__wrapped__.executemany, sql, param_list)

    def _bake_sql(self, sql):
        """
        Method to turn the "sql" argument into a string. Most database backends simply return
        the given object, as it is already a string
        """
        return sql

    def _trace_sql(self, method, sql, params, action=QUERY_ACTION):
        sql_string = self._bake_sql(sql)
        if action == EXEC_ACTION:
            signature = sql_string + "()"
        else:
            signature = self.extract_signature(sql_string)
            action = extract_action_from_signature(signature, action)

        # Truncate sql_string to 10000 characters to prevent large queries from
        # causing an error to APM server.
        sql_string = shorten(sql_string, string_length=10000)

        with capture_span(
            signature,
            span_type="db",
            span_subtype=self.provider_name,
            span_action=action,
            extra={
                "db": {"type": "sql", "statement": sql_string, "instance": getattr(self, "_self_database", None)},
                "destination": self._self_destination_info,
            },
            skip_frames=1,
            leaf=True,
        ) as span:
            if params is None:
                result = method(sql)
            else:
                result = method(sql, params)
            # store "rows affected", but only for DML queries like insert/update/delete
            if span and self.rowcount not in (-1, None) and signature.startswith(self.DML_QUERIES):
                span.update_context("db", {"rows_affected": self.rowcount})
            return result

    def extract_signature(self, sql):
        raise NotImplementedError()


class ConnectionProxy(wrapt.ObjectProxy):
    cursor_proxy = CursorProxy

    def __init__(self, wrapped, destination_info=None) -> None:
        super(ConnectionProxy, self).__init__(wrapped)
        self._self_destination_info = destination_info

    def cursor(self, *args, **kwargs):
        return self.cursor_proxy(self.__wrapped__.cursor(*args, **kwargs), self._self_destination_info)


class DbApi2Instrumentation(AbstractInstrumentedModule):
    connect_method = None

    def call(self, module, method, wrapped, instance, args, kwargs):
        return ConnectionProxy(wrapped(*args, **kwargs))

    def call_if_sampling(self, module, method, wrapped, instance, args, kwargs):
        # Contrasting to the superclass implementation, we *always* want to
        # return a proxied connection, even if there is no ongoing atatus
        # transaction yet. This ensures that we instrument the cursor once
        # the transaction started.
        return self.call(module, method, wrapped, instance, args, kwargs)
